import { memoize, uniqueId } from "lodash-es";
import {
    Background,
    Badge,
    cardSchema2,
    Heading,
    HeadingSize,
} from "@microsoft/fast-components-react-msft";
import {
    neutralLayerL1,
    neutralLayerL2,
    neutralLayerL3,
} from "@microsoft/fast-components-styles-msft";
import {
    Canvas,
    Container,
    Pane,
    PaneResizeDirection,
    Row,
} from "@microsoft/fast-layouts-react";
import Foundation, { HandledProps } from "@microsoft/fast-components-foundation-react";
import { Direction } from "@microsoft/fast-web-utilities";
import React from "react";
import ReactDOM from "react-dom";
import {
    AjvMapper,
    CustomMessageIncomingOutgoing,
    MessageSystem,
    MessageSystemDataTypeAction,
    MessageSystemType,
    SchemaDictionary,
} from "@microsoft/fast-tooling";
import {
    ControlConfig,
    ControlType,
    defaultDevices,
    Device,
    Display,
    LinkedDataControl,
    ModularForm,
    ModularNavigation,
    ModularViewer,
    SelectDevice,
    StandardControlPlugin,
    ViewerCustomAction,
} from "@microsoft/fast-tooling-react";
import {
    ControlContext,
    ControlOnChangeConfig,
} from "@microsoft/fast-tooling-react/dist/form/templates/types";
import FASTMessageSystemWorker from "@microsoft/fast-tooling/dist/message-system.min.js";
import {
    AccentColorPicker,
    Dimension,
    DirectionSwitch,
    ThemeSelector,
} from "@microsoft/site-utilities";
import { fastDesignSystemDefaults } from "@microsoft/fast-components/src/fast-design-system";
import { StandardLuminance } from "@microsoft/fast-components";
import {
    CreatorHandledProps,
    CreatorProps,
    CreatorState,
    ProjectFile,
    ProjectFileView,
} from "./creator.props";
import {
    divTag,
    linkedDataExamples,
    nativeElementSchemas,
    webComponentSchemas,
} from "./configs";
import { ProjectFileTransfer } from "./components";
import { selectDeviceOverrideStyles } from "./utilities/style-overrides";
import { previewReady } from "./preview";
import { textSchema } from "./utilities";
import { fastDesignSystemProviderId } from "./configs/fast-design-system-provider.definition";

const fastMessageSystemWorker = new FASTMessageSystemWorker();
let fastMessageSystem: MessageSystem;
const schemaDictionary: SchemaDictionary = {
    ...webComponentSchemas,
    ...nativeElementSchemas,
    [textSchema.id]: textSchema,
    [webComponentSchemas[fastDesignSystemProviderId].id]:
        webComponentSchemas[fastDesignSystemProviderId],
};

export const previewDirection: string = "PREVIEW::DIRECTION";
export const previewAccentColor: string = "PREVIEW::ACCENTCOLOR";
export const previewTheme: string = "PREVIEW::THEME";

class Creator extends Foundation<CreatorHandledProps, {}, CreatorState> {
    public static displayName: string = "Creator";

    private viewerContainerRef: React.RefObject<HTMLDivElement> = React.createRef();

    private viewerContentAreaPadding: number = 20;

    private devices: Device[];

    private handleDimensionChange: (
        cb: (value: number) => void
    ) => React.ChangeEventHandler<HTMLInputElement> = memoize(
        (cb: (value: number) => void): React.ChangeEventHandler<HTMLInputElement> => {
            return (e: React.ChangeEvent<HTMLInputElement>): void => {
                const value: number = parseInt(e.target.value, 10);

                if (!isNaN(value)) {
                    cb(value);
                }
            };
        }
    );

    constructor(props: CreatorProps) {
        super(props);

        const initialViewId: string = uniqueId("view");
        const componentLinkedDataId: string = "root";

        this.devices = this.getDevices();

        const initialView: ProjectFileView = {
            dataDictionary: [
                {
                    [componentLinkedDataId]: {
                        schemaId: divTag,
                        data: {},
                    },
                },
                componentLinkedDataId,
            ],
        };

        this.state = {
            xCoord: 0,
            yCoord: 0,
            width: 0,
            height: 0,
            deviceId: this.devices[0].id,
            activeView: initialViewId,
            theme: StandardLuminance.LightMode,
            direction: Direction.ltr,
            accentColor: fastDesignSystemDefaults.accentBaseColor,
            views: {
                [initialViewId]: initialView,
            },
            activeDictionaryId: componentLinkedDataId,
            previewReady: false,
        };

        if ((window as any).Worker) {
            fastMessageSystem = new MessageSystem({
                webWorker: fastMessageSystemWorker,
            });
            new AjvMapper({
                messageSystem: fastMessageSystem,
            });
            fastMessageSystem.add({ onMessage: this.handleMessageSystem });
        }
    }

    public render(): React.ReactNode {
        return (
            <Background value={neutralLayerL1}>
                <Container>
                    <Row style={{ flex: "1" }}>
                        <Pane resizable={true} resizeFrom={PaneResizeDirection.east}>
                            <Background
                                value={neutralLayerL3}
                                drawBackground={true}
                                style={{
                                    display: "flex",
                                    height: "32px",
                                    alignItems: "center",
                                    justifyContent: "space-between",
                                    padding: "0 8px",
                                }}
                            >
                                <Heading size={HeadingSize._6}>FAST Creator</Heading>
                                <Badge>ALPHA</Badge>
                            </Background>
                            <ModularNavigation messageSystem={fastMessageSystem} />
                            <ProjectFileTransfer
                                projectFile={this.state}
                                onUpdateProjectFile={this.handleUpdateProjectFile}
                            />
                        </Pane>
                        <Canvas>
                            <Row fill={true}>
                                <Background
                                    value={neutralLayerL2}
                                    drawBackground={true}
                                    style={{
                                        width: "100%",
                                        display: "flex",
                                        alignItems: "center",
                                        justifyContent: "space-between",
                                        padding: "0 8px",
                                    }}
                                >
                                    <SelectDevice
                                        devices={this.devices}
                                        activeDeviceId={this.state.deviceId}
                                        onUpdateDevice={this.handleUpdateDevice}
                                        jssStyleSheet={selectDeviceOverrideStyles}
                                        disabled={!this.state.previewReady}
                                    />
                                    <Dimension
                                        width={this.state.width}
                                        height={this.state.height}
                                        onUpdateWidth={this.handleUpdateWidth}
                                        onUpdateHeight={this.handleUpdateHeight}
                                        onUpdateOrientation={this.handleUpdateOrientation}
                                        onDimensionChange={this.handleDimensionChange}
                                        disabled={!this.state.previewReady}
                                    />
                                    <div
                                        style={{
                                            display: "flex",
                                            marginLeft: "auto",
                                        }}
                                    >
                                        <ThemeSelector
                                            id={"theme-selector"}
                                            theme={this.state.theme}
                                            onUpdateTheme={this.handleUpdateTheme}
                                            disabled={!this.state.previewReady}
                                        />
                                        <DirectionSwitch
                                            id={"direction-switch"}
                                            direction={this.state.direction}
                                            onUpdateDirection={this.handleUpdateDirection}
                                            disabled={!this.state.previewReady}
                                        />
                                        <AccentColorPicker
                                            id={"accent-color-picker"}
                                            accentBaseColor={this.state.accentColor}
                                            onAccentColorPickerChange={
                                                this.handleAccentColorPickerChange
                                            }
                                            disabled={!this.state.previewReady}
                                        />
                                    </div>
                                </Background>
                            </Row>
                            <Row fill={true} style={{ height: "100%" }}>
                                <div
                                    ref={this.viewerContainerRef}
                                    style={{ width: "100%", height: "100%" }}
                                >
                                    <ModularViewer
                                        iframeSrc={"/preview"}
                                        messageSystem={fastMessageSystem}
                                        width={this.state.width}
                                        height={this.state.height}
                                        onUpdateHeight={this.handleUpdateHeight}
                                        onUpdateWidth={this.handleUpdateWidth}
                                        responsive={true}
                                    />
                                </div>
                            </Row>
                        </Canvas>
                        <Pane>
                            <ModularForm
                                messageSystem={fastMessageSystem}
                                controls={[
                                    new StandardControlPlugin({
                                        type: ControlType.linkedData,
                                        context: ControlContext.fill,
                                        control: (
                                            config: ControlConfig
                                        ): React.ReactNode => {
                                            return (
                                                <LinkedDataControl
                                                    {...config}
                                                    onChange={this.handleAddLinkedData(
                                                        config.dataLocation
                                                    )}
                                                />
                                            );
                                        },
                                    }),
                                ]}
                            />
                        </Pane>
                    </Row>
                </Container>
            </Background>
        );
    }

    private handleAddLinkedData = (
        dataLocation: string
    ): ((e: ControlOnChangeConfig) => void) => {
        return (e: ControlOnChangeConfig): void => {
            fastMessageSystem.postMessage({
                type: MessageSystemType.data,
                action: MessageSystemDataTypeAction.addLinkedData,
                dataLocation,
                dictionaryId: this.state.activeDictionaryId,
                index: e.index,
                linkedData: linkedDataExamples[e.value[0].schemaId],
            });
        };
    };

    private handleMessageSystem = (e: MessageEvent): void => {
        const updatedState: Partial<CreatorState> = {};

        if (e.data.type === MessageSystemType.data) {
            updatedState.views = {
                ...this.state.views,
                [this.state.activeView]: {
                    dataDictionary: e.data.dataDictionary,
                },
            };
        }

        if (
            e.data.type === MessageSystemType.custom &&
            e.data.action === ViewerCustomAction.response
        ) {
            if (e.data.value === previewReady) {
                fastMessageSystem.postMessage({
                    type: MessageSystemType.initialize,
                    data: this.state.views[this.state.activeView].dataDictionary,
                    schemaDictionary,
                });
                updatedState.previewReady = true;
            } else if (e.data.value.type === MessageSystemType.navigation) {
                fastMessageSystem.postMessage(e.data.value);
            }
        }

        this.setState(updatedState as CreatorState);
    };

    private handleUpdateProjectFile = (projectFile: ProjectFile): void => {
        this.setState(projectFile, () =>
            fastMessageSystem.postMessage({
                type: MessageSystemType.initialize,
                data: projectFile.views[projectFile.activeView].dataDictionary,
                schemaDictionary,
            })
        );
    };

    public componentDidMount(): void {
        this.setViewerToFullSize();
    }

    private setViewerToFullSize(): void {
        const viewerContainer: HTMLDivElement | null = this.viewerContainerRef.current;

        if (viewerContainer) {
            /* eslint-disable-next-line react/no-find-dom-node */
            const viewerNode: Element | Text | null = ReactDOM.findDOMNode(
                viewerContainer
            );

            if (viewerNode instanceof Element) {
                const height: number =
                    viewerNode.clientHeight - this.viewerContentAreaPadding * 2;
                const width: number =
                    viewerNode.clientWidth - this.viewerContentAreaPadding * 2;
                this.setState({
                    width,
                    height: height - 24, // 24 is height of view label
                });
            }
        }
    }

    private getDevices(): Device[] {
        return defaultDevices.concat({
            id: "desktop",
            displayName: "Desktop (1920x1080)",
            display: Display.fixed,
            width: 1920,
            height: 1080,
        });
    }

    private getDeviceById(id: string): Device | void {
        return this.devices.find((device: Device): boolean => {
            return device.id === id;
        });
    }

    private handleUpdateDevice = (deviceId: string): void => {
        const device: Device | void = this.getDeviceById(deviceId);
        let height: number = this.state.height;
        let width: number = this.state.width;

        if (device) {
            height =
                device.display === Display.responsive
                    ? this.state.height
                    : (device.height as number);
            width =
                device.display === Display.responsive
                    ? this.state.width
                    : (device.width as number);
        }

        this.setState({
            deviceId,
            height,
            width,
        });
    };

    private handleUpdateOrientation = (): void => {
        this.setState({
            width: this.state.height,
            height: this.state.width,
        });
    };

    private setResponsiveDeviceId(): void {
        const activeDevice: Device | void = this.getDeviceById(this.state.deviceId);

        if (activeDevice && activeDevice.display !== Display.responsive) {
            this.setState({
                deviceId: Display.responsive,
            });
        }
    }

    private handleUpdateHeight = (height: number): void => {
        this.setResponsiveDeviceId();
        this.setState({
            height,
        });
    };

    private handleUpdateWidth = (width: number): void => {
        this.setResponsiveDeviceId();
        this.setState({
            width,
        });
    };

    private handleUpdateDirection = (): void => {
        const updatedDirection: Direction =
            this.state.direction === Direction.ltr ? Direction.rtl : Direction.ltr;
        this.setState({
            direction: updatedDirection,
        });

        fastMessageSystem.postMessage({
            type: MessageSystemType.custom,
            id: previewDirection,
            value: updatedDirection,
        } as CustomMessageIncomingOutgoing);
    };

    private handleUpdateTheme = (): void => {
        const value: StandardLuminance =
            this.state.theme === StandardLuminance.LightMode
                ? StandardLuminance.DarkMode
                : StandardLuminance.LightMode;

        this.setState({
            theme: value,
        });

        fastMessageSystem.postMessage({
            type: MessageSystemType.custom,
            id: previewTheme,
            value,
        } as CustomMessageIncomingOutgoing);
    };

    /**
     * Event handler for all color input changes
     */
    private handleAccentColorPickerChange = (
        e: React.FormEvent<HTMLInputElement>
    ): void => {
        const value: string = e.currentTarget.value;

        this.setState({
            accentColor: value,
        });

        fastMessageSystem.postMessage({
            type: MessageSystemType.custom,
            id: previewAccentColor,
            value,
        } as CustomMessageIncomingOutgoing);
    };
}

export default Creator;
